"""Collection of customised padding layers."""

__author__ = 'Felix Kleinert'
__date__ = '2020-03-02'


from typing import Union, Tuple

import numpy as np
import tensorflow as tf
# from tensorflow.keras.backend.common import normalize_data_format
from tensorflow.keras.layers import ZeroPadding2D
# from tensorflow.keras.layers.convolutional import _ZeroPadding
from tensorflow.keras.layers import Layer
# from tensorflow.keras.legacy import interfaces
from mlair.keras_legacy import interfaces
# from tensorflow.keras.utils import conv_utils
from mlair.keras_legacy import conv_utils
# from tensorflow.keras.utils.generic_utils import transpose_shape
# from mlair.keras_legacy.generic_utils import transpose_shape


""" TAKEN FROM KERAS 2.2.0 """
def transpose_shape(shape, target_format, spatial_axes):
    """Converts a tuple or a list to the correct `data_format`.
    It does so by switching the positions of its elements.
    # Arguments
        shape: Tuple or list, often representing shape,
            corresponding to `'channels_last'`.
        target_format: A string, either `'channels_first'` or `'channels_last'`.
        spatial_axes: A tuple of integers.
            Correspond to the indexes of the spatial axes.
            For example, if you pass a shape
            representing (batch_size, timesteps, rows, cols, channels),
            then `spatial_axes=(2, 3)`.
    # Returns
        A tuple or list, with the elements permuted according
        to `target_format`.
    # Example
    ```python
        >>> # from keras.utils.generic_utils import transpose_shape
        >>> transpose_shape((16, 128, 128, 32),'channels_first', spatial_axes=(1, 2))
        (16, 32, 128, 128)
        >>> transpose_shape((16, 128, 128, 32), 'channels_last', spatial_axes=(1, 2))
        (16, 128, 128, 32)
        >>> transpose_shape((128, 128, 32), 'channels_first', spatial_axes=(0, 1))
        (32, 128, 128)
    ```
    # Raises
        ValueError: if `value` or the global `data_format` invalid.
    """
    if target_format == 'channels_first':
        new_values = shape[:spatial_axes[0]]
        new_values += (shape[-1],)
        new_values += tuple(shape[x] for x in spatial_axes)

        if isinstance(shape, list):
            return list(new_values)
        return new_values
    elif target_format == 'channels_last':
        return shape
    else:
        raise ValueError('The `data_format` argument must be one of '
                         '"channels_first", "channels_last". Received: ' +
                         str(target_format))

""" TAKEN FROM KERAS 2.2.0 """
def normalize_data_format(value):
    """Checks that the value correspond to a valid data format.
    # Arguments
        value: String or None. `'channels_first'` or `'channels_last'`.
    # Returns
        A string, either `'channels_first'` or `'channels_last'`
    # Example
    ```python
        >>> from tensorflow.keras import backend as K
        >>> K.normalize_data_format(None)
        'channels_first'
        >>> K.normalize_data_format('channels_last')
        'channels_last'
    ```
    # Raises
        ValueError: if `value` or the global `data_format` invalid.
    """
    if value is None:
        value = 'channels_last'
    data_format = value.lower()
    if data_format not in {'channels_first', 'channels_last'}:
        raise ValueError('The `data_format` argument must be one of '
                         '"channels_first", "channels_last". Received: ' +
                         str(value))
    return data_format


class PadUtils:
    """Helper class for advanced padding."""

    @staticmethod
    def get_padding_for_same(kernel_size: Tuple[int], strides: int = 1) -> Tuple[int]:
        """
        Calculate padding size to keep input and output dimensions equal for a given kernel size.

        .. hint:: custom paddings are currently only implemented for strides = 1

        :param kernel_size: size of padding kernel size
        :param strides: number of strides (default 1, currently only strides=1 supported)

        :return: padding size
        """
        if strides != 1:
            raise NotImplementedError("Strides other than 1 not implemented!")
        if not all(isinstance(k, int) for k in kernel_size):
            raise ValueError(f"The `kernel_size` argument must have a tuple of integers. Got: {kernel_size} "
                             f"of type {[type(k) for k in kernel_size]}")

        ks = np.array(kernel_size, dtype=np.int64)

        if any(k <= 0 for k in ks):
            raise ValueError(f"All values of kernel_size must be > 0. Got: {kernel_size} ")

        if all(k % 2 == 1 for k in ks):  # (d & 0x1 for d in ks):
            pad = ((ks - 1) / 2).astype(np.int64)
            # convert numpy int to base int
            pad = [int(v.item()) for v in pad]
            return tuple(pad)
        else:
            raise NotImplementedError(f"even kernel size not implemented. Got {kernel_size}")

    @staticmethod
    def spatial_2d_padding(padding=((1, 1), (1, 1)), data_format=None):
        """
        Pad the 2nd and 3rd dimensions of a 4D tensor.

        # Arguments
            x: Tensor or variable.
            padding: Tuple of 2 tuples, padding pattern.
            data_format: string, `"channels_last"` or `"channels_first"`.

        # Returns
            A padded 4D tensor.

        # Raises
            ValueError: if `data_format` is neither `"channels_last"` or `"channels_first"`.
        """
        assert len(padding) == 2
        assert len(padding[0]) == 2
        assert len(padding[1]) == 2
        data_format = normalize_data_format(data_format)

        pattern = [[0, 0],
                   list(padding[0]),
                   list(padding[1]),
                   [0, 0]]
        pattern = transpose_shape(pattern, data_format, spatial_axes=(1, 2))
        return pattern

    @staticmethod
    def check_padding_format(padding):
        """Check padding format (int, 1D or 2D, >0)."""
        if isinstance(padding, int):
            normalized_padding = ((padding, padding), (padding, padding))
        elif hasattr(padding, '__len__'):
            if len(padding) != 2:
                raise ValueError('`padding` should have two elements. '
                                 'Found: ' + str(padding))
            for idx_pad, sub_pad in enumerate(padding):
                if isinstance(sub_pad, str):
                    raise ValueError(f'`padding[{idx_pad}]` is str but must be int')
                if hasattr(sub_pad, '__len__'):
                    if len(sub_pad) != 2:
                        raise ValueError(f'`padding[{idx_pad}]` should have one or two elements. '
                                         f'Found: {padding[idx_pad]}')
                    if not all(isinstance(sub_k, int) for sub_k in padding[idx_pad]):
                        raise ValueError(f'`padding[{idx_pad}]` should have one or two elements of type int. '
                                         f"Found:{padding[idx_pad]} of type {[type(sub_k) for sub_k in padding[idx_pad]]}")
            height_padding = conv_utils.normalize_tuple(padding[0], 2,
                                                        '1st entry of padding')
            if not all(k >= 0 for k in height_padding):
                raise ValueError(
                    f"The `1st entry of padding` argument must be >= 0. Received: {padding[0]} of type {type(padding[0])}")
            width_padding = conv_utils.normalize_tuple(padding[1], 2,
                                                       '2nd entry of padding')
            if not all(k >= 0 for k in width_padding):
                raise ValueError(
                    f"The `2nd entry of padding` argument must be >= 0. Received: {padding[1]} of type {type(padding[1])}")
            normalized_padding = (height_padding, width_padding)
        else:
            raise ValueError('`padding` should be either an int, '
                             'a tuple of 2 ints '
                             '(symmetric_height_pad, symmetric_width_pad), '
                             'or a tuple of 2 tuples of 2 ints '
                             '((top_pad, bottom_pad), (left_pad, right_pad)). '
                             f'Found: {padding} of type {type(padding)}')
        return normalized_padding

""" TAKEN FROM KERAS 2.2.0 """
class InputSpec(object):
    """Specifies the ndim, dtype and shape of every input to a layer.
    Every layer should expose (if appropriate) an `input_spec` attribute:
    a list of instances of InputSpec (one per input tensor).
    A None entry in a shape is compatible with any dimension,
    a None shape is compatible with any shape.
    # Arguments
        dtype: Expected datatype of the input.
        shape: Shape tuple, expected shape of the input
            (may include None for unchecked axes).
        ndim: Integer, expected rank of the input.
        max_ndim: Integer, maximum rank of the input.
        min_ndim: Integer, minimum rank of the input.
        axes: Dictionary mapping integer axes to
            a specific dimension value.
    """

    def __init__(self, dtype=None,
                 shape=None,
                 ndim=None,
                 max_ndim=None,
                 min_ndim=None,
                 axes=None):
        self.dtype = dtype
        self.shape = shape
        if shape is not None:
            self.ndim = len(shape)
        else:
            self.ndim = ndim
        self.max_ndim = max_ndim
        self.min_ndim = min_ndim
        self.axes = axes or {}

    def __repr__(self):
        spec = [('dtype=' + str(self.dtype)) if self.dtype else '',
                ('shape=' + str(self.shape)) if self.shape else '',
                ('ndim=' + str(self.ndim)) if self.ndim else '',
                ('max_ndim=' + str(self.max_ndim)) if self.max_ndim else '',
                ('min_ndim=' + str(self.min_ndim)) if self.min_ndim else '',
                ('axes=' + str(self.axes)) if self.axes else '']
        return 'InputSpec(%s)' % ', '.join(x for x in spec if x)

""" TAKEN FROM KERAS 2.2.0 """
class _ZeroPadding(Layer):
    """Abstract nD ZeroPadding layer (private, used as implementation base).
    # Arguments
        padding: Tuple of tuples of two ints. Can be a tuple of ints when
            rank is 1.
        data_format: A string,
            one of `"channels_last"` or `"channels_first"`.
            The ordering of the dimensions in the inputs.
            `"channels_last"` corresponds to inputs with shape
            `(batch, ..., channels)` while `"channels_first"` corresponds to
            inputs with shape `(batch, channels, ...)`.
            It defaults to the `image_data_format` value found in your
            Keras config file at `~/.keras/keras.json`.
            If you never set it, then it will be "channels_last".
    """
    def __init__(self, padding, data_format=None, **kwargs):
        # self.rank is 1 for ZeroPadding1D, 2 for ZeroPadding2D.
        self.rank = len(padding)
        self.padding = padding
        self.data_format = normalize_data_format(data_format)
        self.input_spec = tf.keras.layers.InputSpec(ndim=self.rank + 2)
        super(_ZeroPadding, self).__init__(**kwargs)

    def call(self, inputs):
        raise NotImplementedError

    def compute_output_shape(self, input_shape):
        padding_all_dims = ((0, 0),) + self.padding + ((0, 0),)
        spatial_axes = list(range(1, 1 + self.rank))
        padding_all_dims = transpose_shape(padding_all_dims,
                                           self.data_format,
                                           spatial_axes)
        output_shape = list(input_shape)
        for dim in range(len(output_shape)):
            if output_shape[dim] is not None:
                output_shape[dim] += sum(padding_all_dims[dim])
        return tuple(output_shape)

    def get_config(self):
        config = {'padding': self.padding,
                  'data_format': self.data_format}
        base_config = super(_ZeroPadding, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))


class ReflectionPadding2D(_ZeroPadding):
    """
    Reflection padding layer for 2D input.

    This custom padding layer is built on keras' zero padding layers. Doc is copy and pasted from the original
    functions/methods:

    This layer can add rows and columns of reflected values
    at the top, bottom, left and right side of an image like tensor.

    Example:
                                                        6, 5,  4, 5, 6,  5, 4
                                                              _________
     1, 2, 3     RefPad(padding=[[1, 1,], [2, 2]])      3, 2,| 1, 2, 3,| 2, 1
     4, 5, 6     =============================>>>>      6, 5,| 4, 5, 6,| 5, 4
                                                              _________
                                                        3, 2,  1, 2, 3,  2, 1



    # Arguments
        padding: int, or tuple of 2 ints, or tuple of 2 tuples of 2 ints.
            - If int: the same symmetric padding
                is applied to height and width.
            - If tuple of 2 ints:
                interpreted as two different
                symmetric padding values for height and width:
                `(symmetric_height_pad, symmetric_width_pad)`.
            - If tuple of 2 tuples of 2 ints:
                interpreted as
                `((top_pad, bottom_pad), (left_pad, right_pad))`
        data_format: A string,
            one of `"channels_last"` or `"channels_first"`.
            The ordering of the dimensions in the inputs.
            `"channels_last"` corresponds to inputs with shape
            `(batch, height, width, channels)` while `"channels_first"`
            corresponds to inputs with shape
            `(batch, channels, height, width)`.
            It defaults to the `image_data_format` value found in your
            Keras config file at `~/.keras/keras.json`.
            If you never set it, then it will be "channels_last".

    # Input shape
        4D tensor with shape:
        - If `data_format` is `"channels_last"`:
            `(batch, rows, cols, channels)`
        - If `data_format` is `"channels_first"`:
            `(batch, channels, rows, cols)`

    # Output shape
        4D tensor with shape:
        - If `data_format` is `"channels_last"`:
            `(batch, padded_rows, padded_cols, channels)`
        - If `data_format` is `"channels_first"`:
            `(batch, channels, padded_rows, padded_cols)`
    '
    """

    @interfaces.legacy_zeropadding2d_support
    def __init__(self,
                 padding=(1, 1),
                 data_format=None,
                 **kwargs):
        """Initialise ReflectionPadding2D."""
        normalized_padding = PadUtils.check_padding_format(padding=padding)
        super(ReflectionPadding2D, self).__init__(normalized_padding,
                                                  data_format,
                                                  **kwargs)

    def call(self, inputs, mask=None):
        """Call ReflectionPadding2D."""
        pattern = PadUtils.spatial_2d_padding(padding=self.padding, data_format=self.data_format)
        return tf.pad(tensor=inputs, paddings=pattern, mode='REFLECT')


class SymmetricPadding2D(_ZeroPadding):
    """
    Symmetric padding layer for 2D input.

    This custom padding layer is built on keras' zero padding layers. Doc is copy pasted from the original
    functions/methods:

    This layer can add rows and columns of symmetric values
    at the top, bottom, left and right side of an image like tensor.

        Example:
                                                        2, 1,  1, 2, 3,  3, 2
                                                              _________
     1, 2, 3     SymPad(padding=[[1, 1,], [2, 2]])      2, 1,| 1, 2, 3,| 3, 2
     4, 5, 6     =============================>>>>      5, 4,| 4, 5, 6,| 6, 5
                                                              _________
                                                        5, 4,  4, 5, 6,  6, 5


    '# Arguments
        padding: int, or tuple of 2 ints, or tuple of 2 tuples of 2 ints.
            - If int: the same symmetric padding
                is applied to height and width.
            - If tuple of 2 ints:
                interpreted as two different
                symmetric padding values for height and width:
                `(symmetric_height_pad, symmetric_width_pad)`.
            - If tuple of 2 tuples of 2 ints:
                interpreted as
                `((top_pad, bottom_pad), (left_pad, right_pad))`
        data_format: A string,
            one of `"channels_last"` or `"channels_first"`.
            The ordering of the dimensions in the inputs.
            `"channels_last"` corresponds to inputs with shape
            `(batch, height, width, channels)` while `"channels_first"`
            corresponds to inputs with shape
            `(batch, channels, height, width)`.
            It defaults to the `image_data_format` value found in your
            Keras config file at `~/.keras/keras.json`.
            If you never set it, then it will be "channels_last".

    # Input shape
        4D tensor with shape:
        - If `data_format` is `"channels_last"`:
            `(batch, rows, cols, channels)`
        - If `data_format` is `"channels_first"`:
            `(batch, channels, rows, cols)`

    # Output shape
        4D tensor with shape:
        - If `data_format` is `"channels_last"`:
            `(batch, padded_rows, padded_cols, channels)`
        - If `data_format` is `"channels_first"`:
            `(batch, channels, padded_rows, padded_cols)`
    '
    """

    @interfaces.legacy_zeropadding2d_support
    def __init__(self,
                 padding=(1, 1),
                 data_format=None,
                 **kwargs):
        """Initialise SymmetricPadding2D."""
        normalized_padding = PadUtils.check_padding_format(padding=padding)
        super(SymmetricPadding2D, self).__init__(normalized_padding,
                                                 data_format,
                                                 **kwargs)

    def call(self, inputs, mask=None):
        """Call SymmetricPadding2D."""
        pattern = PadUtils.spatial_2d_padding(padding=self.padding, data_format=self.data_format)
        return tf.pad(tensor=inputs, paddings=pattern, mode='SYMMETRIC')


class Padding2D:
    """
    Combine all implemented padding methods.

    You can call this method by defining a specific padding type. The __call__ method will return the corresponding
    Padding layer.

    .. code-block:: python

        input_x = ... #  your input data
        kernel_size = (5, 1)
        padding_size = PadUtils.get_padding_for_same(kernel_size)

        tower = layers.Conv2D(...)(input_x)
        tower = layers.Activation(...)(tower)
        tower = Padding2D('ZeroPad2D')(padding=padding_size, name=f'Custom_Pad')(tower)

    Padding type can either be set by a string or directly by using an instance of a valid padding class.
    """

    allowed_paddings = {
        **dict.fromkeys(("RefPad2D", "ReflectionPadding2D"), ReflectionPadding2D),
        **dict.fromkeys(("SymPad2D", "SymmetricPadding2D"), SymmetricPadding2D),
        **dict.fromkeys(("ZeroPad2D", "ZeroPadding2D"), ZeroPadding2D)
    }
    padding_type = Union[ReflectionPadding2D, SymmetricPadding2D, ZeroPadding2D]

    def __init__(self, padding_type: Union[str, padding_type]):
        """Set padding type."""
        self.padding_type = padding_type

    def _check_and_get_padding(self):
        if isinstance(self.padding_type, str):
            try:
                pad2d = self.allowed_paddings[self.padding_type]
            except KeyError as e:
                raise NotImplementedError(
                    f"`{e}' is not implemented as padding. Use one of those: i) `RefPad2D', ii) `SymPad2D', "
                    f"iii) `ZeroPad2D'")
        else:
            if self.padding_type in self.allowed_paddings.values():
                pad2d = self.padding_type
            else:
                raise TypeError(f"`{self.padding_type.__name__}' is not a valid padding layer type. "
                                "Use one of those: "
                                "i) ReflectionPadding2D, ii) SymmetricPadding2D, iii) ZeroPadding2D")
        return pad2d

    def __call__(self, *args, **kwargs):
        """Call padding."""
        return self._check_and_get_padding()(*args, **kwargs)


if __name__ == '__main__':
    from tensorflow.keras.models import Model
    from tensorflow.keras.layers import Conv2D, Flatten, Dense, Input

    kernel_1 = (3, 3)
    kernel_2 = (5, 5)
    kernel_3 = (3, 3)
    x = np.array(range(2000)).reshape((-1, 10, 10, 1))
    y = x.mean(axis=(1, 2))

    x_input = Input(shape=x.shape[1:])
    pad1 = PadUtils.get_padding_for_same(kernel_size=kernel_1)
    x_out = ReflectionPadding2D(padding=pad1, name="RefPAD")(x_input)
    x_out = Conv2D(5, kernel_size=kernel_1, activation='relu')(x_out)

    pad2 = PadUtils.get_padding_for_same(kernel_size=kernel_2)
    x_out = SymmetricPadding2D(padding=pad2, name="SymPAD")(x_out)
    x_out = Conv2D(2, kernel_size=kernel_2, activation='relu')(x_out)

    pad3 = PadUtils.get_padding_for_same(kernel_size=kernel_3)
    x_out = Padding2D('RefPad2D')(padding=pad3, name="Padding2D_RefPad")(x_out)
    x_out = Conv2D(2, kernel_size=kernel_3, activation='relu')(x_out)
    x_out = Flatten()(x_out)
    x_out = Dense(1, activation='linear')(x_out)

    model = Model(inputs=x_input, outputs=x_out)
    model.compile('adam', loss='mse')
    model.summary()
    model.fit(x, y, epochs=10)
